#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "resource.h"

/* The padding calculation assumes this is a power of two. */
#define PAGE_SIZE		64

HINSTANCE instance;
HANDLE ser_handle = INVALID_HANDLE_VALUE;
HWND main_win;
volatile HANDLE upload_thread = INVALID_HANDLE_VALUE;
volatile int ser_need_abort = 0;
unsigned char *upload_data = 0;

void upload_shutdown(void);

void browse()
{
	OPENFILENAME ofn;
	char filename[256];

	memset(&ofn, 0, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = main_win;
	ofn.hInstance = instance;
	ofn.lpstrFile = filename;
	ofn.nMaxFile = sizeof(filename);
	ofn.lpstrFilter = "Binaries (*.bin)\0*.bin\0";

	filename[0] = 0;

	if (GetOpenFileName(&ofn))
		SetDlgItemText(main_win, IDC_FILENAME, filename);
}

int ser_init(const char *port)
{
	DCB dcb;
	COMMTIMEOUTS timeouts;

	if (ser_handle != INVALID_HANDLE_VALUE)
		return 1;

	ser_handle = CreateFile(port, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, 0);
	if (ser_handle == INVALID_HANDLE_VALUE)
		return 0;

	memset(&timeouts, 0, sizeof(timeouts));
	timeouts.ReadIntervalTimeout = MAXDWORD;
	timeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
	timeouts.ReadTotalTimeoutConstant = 10;
	timeouts.WriteTotalTimeoutConstant = MAXDWORD;
	timeouts.WriteTotalTimeoutMultiplier = MAXDWORD;
	if (!SetCommTimeouts(ser_handle, &timeouts))
	{
		ser_shutdown();
		return 0;
	}

	dcb.DCBlength = sizeof(dcb);
	if (!GetCommState(ser_handle, &dcb))
	{
		ser_shutdown();
		return 0;
	}

	dcb.BaudRate = CBR_115200;
	dcb.ByteSize = 8;
	dcb.StopBits = ONESTOPBIT;
	dcb.Parity = NOPARITY;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.fBinary = TRUE;

	if (!SetCommState(ser_handle, &dcb))
	{
		ser_shutdown();
		return 0;
	}

	return 1;
}

void ser_shutdown()
{
	if (ser_handle != INVALID_HANDLE_VALUE)
	{
		CloseHandle(ser_handle);
		ser_handle = INVALID_HANDLE_VALUE;
	}
}

void ser_set_dtr()
{
	if (ser_handle != INVALID_HANDLE_VALUE)
		EscapeCommFunction(ser_handle, SETDTR);
}

void ser_clear_dtr()
{
	if (ser_handle != INVALID_HANDLE_VALUE)
		EscapeCommFunction(ser_handle, CLRDTR);
}

unsigned char ser_receive()
{
	unsigned char ch;
	DWORD num;

	/* Try to read until either a byte arrives or the thread is terminated.
	 * This is gross and is necessary because, as far as I can tell, there is
	 * no way to cancel an I/O operation in another thread in Windows and
	 * overlapped I/O isn't working (it blocks).
	 */
	while (!ser_need_abort)
	{
		num = 0;
		if (ReadFile(ser_handle, &ch, 1, &num, NULL) && num == 1)
			return ch;
	}

	/* Terminate this thread (only the uploader thread is allowed to do serial I/O).
	 * Cleanup will be done in upload_abort(), since updating controls here would deadlock
	 * (it looks like the dialog event loop has a mutex that is needed to update the controls,
	 * and the GUI thread will block waiting for this thread to terminate).
	 */
	ExitThread(0);

	return 0;
}

void ser_send(unsigned char ch)
{
	DWORD num;

	WriteFile(ser_handle, &ch, 1, &num, NULL);
}

/* Stops the uploader thread at the next blocking receive.
 * This may only be called by the GUI thread.
 */
void upload_abort()
{
	HWND status = GetDlgItem(main_win, IDC_STATUS);
	HWND progress = GetDlgItem(main_win, IDC_PROGRESS);
	HWND button = GetDlgItem(main_win, IDC_UPLOAD);
	HANDLE thread = upload_thread;

	SetWindowText(status, "Stopping...");
	ShowWindow(status, SW_SHOW);
	ShowWindow(progress, SW_HIDE);
	EnableWindow(button, FALSE);

	ser_need_abort = 1;

	/* Wait for the thread to terminate */
	if (WaitForSingleObject(thread, 1000) != WAIT_OBJECT_0)
	{
		MessageBox(main_win, "Uploader thread failed to terminate - Tell Ben and restart the program.",
			"RJTE Uploader - Oops", MB_ICONSTOP | MB_OK);
		ExitProcess(0);
	}
	ser_need_abort = 0;

	EnableWindow(button, TRUE);

	/* Do cleanup here, rather than in the uploader thread (as would normally be done).
	 * See ser_receive() for why.
	 */
	upload_shutdown();

	SetWindowText(status, "Aborted!");
	SetFocus(button);
}

/* Cleans up and terminates the uploader thread */
void upload_shutdown()
{
	ser_clear_dtr();
	ser_shutdown();
	if (upload_data)
	{
		free(upload_data);
		upload_data = 0;
	}

	upload_thread = INVALID_HANDLE_VALUE;
	SetDlgItemText(main_win, IDC_UPLOAD, "Upload");
}

UINT CALLBACK upload(void *param)
{
	char port[16], filename[256];
	int raw_len, padded_len, i, offset;
	FILE *fp;
	HWND status, progress;
	unsigned char ch;

	upload_data = NULL;
	status = GetDlgItem(main_win, IDC_STATUS);
	progress = GetDlgItem(main_win, IDC_PROGRESS);

	/* Read the file */
	GetDlgItemText(main_win, IDC_FILENAME, filename, sizeof(filename));
	if (!filename[0])
	{
		MessageBox(main_win, "You must select a file to upload first.", "RJTE - Error", MB_ICONERROR | MB_OK);
		upload_shutdown();
		return 0;
	}

	fp = fopen(filename, "rb");
	if (!fp)
	{
		MessageBox(main_win, "Can't read from the specified file.", "RJTE - Error", MB_ICONERROR | MB_OK);
		upload_shutdown();
		return 0;
	}

	fseek(fp, 0, SEEK_END);
	raw_len = ftell(fp);
	/* Round up: This assumes that PAGE_SIZE is a power of two */
	padded_len = (raw_len + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1);
	fseek(fp, 0, SEEK_SET);

	upload_data = malloc(padded_len);
	if (!upload_data)
	{
		MessageBox(main_win, "Out of memory for data.", "RJTE - Error", MB_ICONERROR | MB_OK);
		upload_shutdown();
		return 0;
	}
	memset(upload_data, 0, padded_len);

	fread(upload_data, raw_len, 1, fp);

	fclose(fp);

	/* Open the serial port */
	GetDlgItemText(main_win, IDC_PORT, port, sizeof(port));
	if (!ser_init(port))
	{
		MessageBox(main_win, "Can't open the serial port.", "RJTE - Error", MB_ICONERROR | MB_OK);
		upload_shutdown();
		return 0;
	}

	ser_set_dtr();

	/* Handshake (to confirm that the bootloader is ready to receive data) */
	SetWindowText(status, "Reset the board...");
	ShowWindow(status, SW_SHOW);

	do
	{
		/* Wait for the first handshake byte */
		do
		{
			ch = ser_receive();
		} while (ch != 0x54);

		/* Wait for the second handshake byte */
		ch = ser_receive();
	} while (ch != 0x31);

	/* Return the handshake */
	ser_send(0x48);
	ser_send(0x01);

	/* Show the progress bar */
	SendMessage(progress, PBM_SETRANGE32, 0, padded_len);
	SendMessage(progress, PBM_SETPOS, 0, 0);
	ShowWindow(status, SW_HIDE);
	ShowWindow(progress, SW_SHOW);

	/* Send pages */
	for (offset = 0; offset < padded_len;)
	{
		int checksum = 0;
		unsigned char cl, ch;

		/* Send the data */
		for (i = 0; i < PAGE_SIZE; i += 2)
		{
			unsigned char bl = upload_data[offset++];
			unsigned char bh = upload_data[offset++];

			checksum += bl | (bh << 8);
			checksum &= 0xffff;
			ser_send(bl);
			ser_send(bh);
		}
		SendMessage(progress, PBM_SETPOS, offset, 0);

		/* Get the checksum */
		cl = ser_receive();
		ch = ser_receive();
		if (checksum != (cl | (ch << 8)))
		{
			ShowWindow(progress, SW_HIDE);
			ShowWindow(status, SW_SHOW);
			SetWindowText(status, "Bad checksum!");

			MessageBox(main_win, "Checksum failed - try again", "RJTE Uploader - Error", MB_ICONSTOP | MB_OK);

			upload_shutdown();
			return 0;
		}
	}

	/* Show status text */
	ShowWindow(progress, SW_HIDE);
	ShowWindow(status, SW_SHOW);
	SetWindowText(status, "Done.");

	upload_shutdown();

	return 0;
}

int CALLBACK mainProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
			main_win = hwnd;
			SetDlgItemText(hwnd, IDC_PORT, "COM1");
			return TRUE;
		case WM_CLOSE:
			EndDialog(hwnd, 0);
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDC_UPLOAD:
					if (upload_thread == INVALID_HANDLE_VALUE)
					{
						SetDlgItemText(hwnd, IDC_UPLOAD, "Stop");
						upload_thread = CreateThread(NULL, 0, upload, NULL, 0, NULL);
					} else {
						upload_abort();
						upload_thread = INVALID_HANDLE_VALUE;
						SetDlgItemText(hwnd, IDC_UPLOAD, "Upload");
					}
					break;
				case IDC_BROWSE:
					browse();
					break;
			}
			return TRUE;
	}

	return FALSE;
}

int PASCAL WinMain(HINSTANCE inst, HINSTANCE prevInst, LPSTR cmdLIne, int showCmd)
{
	instance = inst;

	/* Need this for the progress bar */
	InitCommonControls();

	return DialogBox(inst, "Main", NULL, mainProc);
}
